# frozen_string_literal: true

require 'erratum'
require 'json/jwt'
require 'apiphobic/errors/expired'
require 'apiphobic/errors/future_issuance'
require 'apiphobic/errors/invalid_attribute'
require 'apiphobic/errors/unavailable'
require 'apiphobic/tokens/configurable'
require 'apiphobic/tokens/json_web_tokens/null'
require 'apiphobic/tokens/role_predicable'

# rubocop:disable Metrics/ClassLength
module  Apiphobic
module  Tokens
class   JsonWebToken
  include Configurable
  include RolePredicable

  TRANSFORMATION_EXCEPTIONS = [
                                JSON::JWT::Exception,
                                JSON::JWT::InvalidFormat,
                                JSON::JWT::VerificationFailed,
                                JSON::JWT::UnexpectedAlgorithm,
                                OpenSSL::PKey::RSAError,
                                OpenSSL::Cipher::CipherError,
                              ].freeze

  attr_accessor :data,
                :headers,
                :private_key

  def initialize(data:,
                 headers:     {},
                 private_key: configuration.private_key)

    self.data        = data
    self.headers     = headers
    self.private_key = private_key
  end

  def self.build_from_request(request_token)
    return Tokens::JsonWebTokens::Null.instance unless request_token

    data, headers = *request_token

    new(data: data, headers: headers)
  end

  # rubocop:disable Metrics/ParameterLists, Metrics/LineLength
  def self.build(id:                SecureRandom.uuid,
                 audience:          configuration.default_audience,
                 audience_id:,
                 expiration:        Time.now.utc.to_i + (60 * configuration.default_expiration_in_minutes),
                 issuer:            configuration.default_issuer,
                 issued_at:         Time.now.utc,
                 not_before:        Time.now.utc,
                 roles:             configuration.default_roles,
                 subject:           configuration.default_subject,
                 subject_id:        nil,
                 token_private_key: configuration.private_key)

    new(
      private_key: token_private_key,
      data:        {
        'aid' => audience_id,
        'aud' => audience,
        'exp' => expiration.to_i,
        'iat' => issued_at.to_i,
        'iss' => issuer,
        'jti' => id,
        'nbf' => not_before.to_i,
        'rol' => roles.join(','),
        'sid' => subject_id,
        'sub' => subject,
      },
    )
  end
  # rubocop:enable Metrics/ParameterLists, Metrics/LineLength

  def available?
    not_before <= (Time.now.to_i + configuration.default_availability_leeway_in_seconds)
  end

  def blank?
    data.empty?
  end

  def present?
    data.any?
  end

  def empty?
    data.empty?
  end

  def expired?
    expiration <= (Time.now.to_i - configuration.default_expiration_leeway_in_seconds)
  end

  def future_issuance?
    issued_at > Time.now.to_f
  end

  def valid?
    validate!
  rescue Erratum::Errors::InvalidToken
    false
  end

  def valid_audience?
    configuration.available_audiences.empty? ||
    configuration.available_audiences.include?(audience)
  end

  def valid_id?
    id =~ /[a-f0-9]{8}\-[a-f0-9]{4}\-[a-f0-9]{4}\-[a-f0-9]{4}\-[a-f0-9]{12}/i
  end

  def valid_issuer?
    configuration.available_issuers.empty? ||
    configuration.available_issuers.include?(issuer)
  end

  def valid_role?
    roles.empty?                         ||
    configuration.available_roles.empty? ||
    (configuration.available_roles & roles).any?
  end

  def valid_subject?
    configuration.available_subjects.empty? ||
    configuration.available_subjects.include?(subject)
  end

  def validate!
    validate_id!
    validate_availability!
    validate_expiration!
    validate_issuance!
    validate_issuer!
    validate_role!
    validate_subject!

    true
  end

  def audience
    data['aud']
  end

  def audience_id
    data['aid']
  end

  def issued_at
    data['iat']
  end

  def issuer
    data['iss']
  end

  def expiration
    data['exp']
  end

  def expiration_date
    Time.at(expiration)
  end

  def id
    data['jti']
  end

  def not_before
    data['nbf']
  end

  def not_before_date
    Time.at(not_before)
  end

  def subject_id
    data['sid']
  end

  def subject
    data['sub']
  end

  def roles
    @roles ||= data.fetch('rol', '').split(',')
  end

  def to_h
    [data, headers]
  end

  def to_jwt
    @to_jwt ||= JSON::JWT.new(data)
  end

  def to_jwt_s
    @to_jwt_s ||= to_jwt.to_s
  end

  def to_jws
    @to_jws ||= to_jwt.sign(private_key, 'RS256')
  end

  def to_jws_s
    @to_jws_s ||= to_jws.to_s
  end

  def to_jwe
    @to_jwe ||= to_jws.encrypt(private_key, 'RSA-OAEP', 'A256GCM')
  end

  def to_jwe_s
    @to_jwe_s ||= to_jwe.to_s
  end

  def self.from_jwe(encrypted_token,
                    private_key: configuration.private_key)

    return JsonWebTokens::Null.instance if encrypted_token.to_s == ''

    decrypted_token = JSON::JWT
                        .decode(encrypted_token, private_key)
                        .plain_text

    from_jws(decrypted_token, private_key: private_key)
  rescue *TRANSFORMATION_EXCEPTIONS => error
    raise ::Erratum::Errors::InvalidToken.wrap(error)
  end

  def self.from_jws(signed_token,
                    private_key: configuration.private_key)

    return JsonWebTokens::Null.instance if signed_token.to_s == ''

    decoded = JSON::JWT.decode(signed_token, private_key)

    new(data:        decoded.to_h,
        headers:     decoded.header,
        private_key: private_key)
  rescue *TRANSFORMATION_EXCEPTIONS => error
    raise ::Erratum::Errors::InvalidToken.wrap(error)
  end

  private

  def validate_audience!
    return true unless audience
    return true if     valid_audience?

    Erratum.fail Errors::InvalidAttribute,
                 attribute_name:  'audience',
                 attribute_value: audience
  end

  def validate_availability!
    return true unless not_before
    return true if     available?

    Erratum.fail Errors::Unavailable,
                 available_at: not_before_date
  end

  def validate_expiration!
    return true unless expiration
    return true unless expired?

    Erratum.fail Errors::Expired,
                 expired_at: expiration_date
  end

  def validate_id!
    return true unless id
    return true if     valid_id?

    Erratum.fail Errors::InvalidAttribute,
                 attribute_name:     'id',
                 attribute_value:    id,
                 additional_message: 'It must be in the form of a UUID.'
  end

  def validate_issuance!
    return true unless issued_at
    return true unless future_issuance?

    Erratum.fail Errors::FutureIssuance,
                 issued_at: issued_at_date
  end

  def validate_issuer!
    return true unless issuer
    return true if     valid_issuer?

    Erratum.fail Errors::InvalidAttribute,
                 attribute_name:  'issuer',
                 attribute_value: issuer
  end

  def validate_role!
    return true unless roles
    return true if     valid_role?

    Erratum.fail Errors::InvalidAttribute,
                 attribute_name:  'roles',
                 attribute_value: roles
  end

  def validate_subject!
    return true unless subject
    return true if     valid_subject?

    Erratum.fail Errors::InvalidAttribute,
                 attribute_name:  'subject',
                 attribute_value: subject
  end
end
end
end
# rubocop:enable Metrics/ClassLength
